@@ -26,7 +26,6 @@ import sys
 import threading
 import uuid
 
-from oslo_utils import importutils
 import six
 
 if os.name == 'nt':
@@ -37,6 +36,7 @@ else:
 from oslo_utils import encodeutils
 from oslo_utils import strutils
 import prettytable
+import wrapt
 
 from glanceclient._i18n import _
 from glanceclient import exc
@@ -122,7 +122,7 @@ def schema_args(schema_getter, omit=None):
                                                                  kwargs))
         else:
             properties = schema.get('properties', {})
-            for name, property in six.iteritems(properties):
+            for name, property in properties.items():
                 if name in omit:
                     continue
                 param = '--' + name.replace('_', '-')
@@ -177,6 +177,12 @@ def pretty_choice_list(l):
 
 
 def print_list(objs, fields, formatters=None, field_settings=None):
+    '''Prints a list of objects.
+
+    @param objs: Objects to print
+    @param fields: Fields on each object to be printed
+    @param formatters: Custom field formatters
+    '''
     formatters = formatters or {}
     field_settings = field_settings or {}
     pt = prettytable.PrettyTable([f for f in fields], caching=False)
@@ -186,7 +192,7 @@ def print_list(objs, fields, formatters=None, field_settings=None):
         row = []
         for field in fields:
             if field in field_settings:
-                for setting, value in six.iteritems(field_settings[field]):
+                for setting, value in field_settings[field].items():
                     setting_dict = getattr(pt, setting)
                     setting_dict[field] = value
 
@@ -196,16 +202,41 @@ def print_list(objs, fields, formatters=None, field_settings=None):
                 field_name = field.lower().replace(' ', '_')
                 data = getattr(o, field_name, None) or ''
                 row.append(data)
+        count = 0
+        # Converts unicode values in list to string
+        for part in row:
+            count = count + 1
+            if isinstance(part, list):
+                part = unicode_key_value_to_string(part)
+                row[count - 1] = part
         pt.add_row(row)
 
     print(encodeutils.safe_decode(pt.get_string()))
 
 
+def _encode(src):
+    """remove extra 'u' in PY2."""
+    if six.PY2 and isinstance(src, unicode):
+        return src.encode('utf-8')
+    return src
+
+
+def unicode_key_value_to_string(src):
+    """Recursively converts dictionary keys to strings."""
+    if isinstance(src, dict):
+        return dict((_encode(k),
+                    _encode(unicode_key_value_to_string(v)))
+                    for k, v in src.items())
+    if isinstance(src, list):
+        return [unicode_key_value_to_string(l) for l in src]
+    return _encode(src)
+
+
 def print_dict(d, max_column_width=80):
     pt = prettytable.PrettyTable(['Property', 'Value'], caching=False)
     pt.align = 'l'
     pt.max_width = max_column_width
-    for k, v in six.iteritems(d):
+    for k, v in d.items():
         if isinstance(v, (dict, list)):
             v = json.dumps(v)
         pt.add_row([k, v])
@@ -246,21 +277,6 @@ def find_resource(manager, name_or_id):
         return matches[0]
 
 
-def skip_authentication(f):
-    """Function decorator used to indicate a caller may be unauthenticated."""
-    f.require_authentication = False
-    return f
-
-
-def is_authentication_required(f):
-    """Checks to see if the function requires authentication.
-
-    Use the skip_authentication decorator to indicate a caller may
-    skip the authentication step.
-    """
-    return getattr(f, 'require_authentication', True)
-
-
 def env(*vars, **kwargs):
     """Search for the first defined of possibly many env vars.
 
@@ -274,13 +290,6 @@ def env(*vars, **kwargs):
     return kwargs.get('default', '')
 
 
-def import_versioned_module(version, submodule=None):
-    module = 'glanceclient.v%s' % version
-    if submodule:
-        module = '.'.join((module, submodule))
-    return importutils.import_module(module)
-
-
 def exit(msg='', exit_code=1):
     if msg:
         print_err(msg)
@@ -298,10 +307,10 @@ def save_image(data, path):
     :param path: path to save the image to
     """
     if path is None:
-        if six.PY3:
-            image = sys.stdout.buffer
-        else:
-            image = sys.stdout
+        # NOTE(kragniz): for py3 compatibility: sys.stdout.buffer is only
+        # present on py3, otherwise fall back to sys.stdout
+        image = getattr(sys.stdout, 'buffer',
+                        sys.stdout)
     else:
         image = open(path, 'wb')
     try:
@@ -333,7 +342,7 @@ def get_file_size(file_obj):
     """Analyze file-like object and attempt to determine its size.
 
     :param file_obj: file-like object.
-    :retval The file's size or None if it cannot be determined.
+    :retval: The file's size or None if it cannot be determined.
     """
     if (hasattr(file_obj, 'seek') and hasattr(file_obj, 'tell') and
             (six.PY2 or six.PY3 and file_obj.seekable())):
@@ -375,9 +384,12 @@ def get_data_file(args):
             return None
         if not sys.stdin.isatty():
             # (2) image data is provided through standard input
+            image = sys.stdin
+            if hasattr(sys.stdin, 'buffer'):
+                image = sys.stdin.buffer
             if msvcrt:
-                msvcrt.setmode(sys.stdin.fileno(), os.O_BINARY)
-            return sys.stdin
+                msvcrt.setmode(image.fileno(), os.O_BINARY)
+            return image
         else:
             # (3) no image data provided
             return None
@@ -399,7 +411,7 @@ def strip_version(endpoint):
     (scheme, netloc, path, __, __, __) = url_parts
     path = path.lstrip('/')
     # regex to match 'v1' or 'v2.0' etc
-    if re.match('v\d+\.?\d*', path):
+    if re.match(r'v\d+\.?\d*', path):
         version = float(path.lstrip('v'))
         endpoint = scheme + '://' + netloc
     return endpoint, version
@@ -407,8 +419,10 @@ def strip_version(endpoint):
 
 def print_image(image_obj, human_readable=False, max_col_width=None):
     ignore = ['self', 'access', 'file', 'schema']
-    image = dict([item for item in six.iteritems(image_obj)
+    image = dict([item for item in image_obj.items()
                   if item[0] not in ignore])
+    if 'virtual_size' in image:
+        image['virtual_size'] = image.get('virtual_size') or 'Not available'
     if human_readable:
         image['size'] = make_size_human_readable(image['size'])
     if str(max_col_width).isdigit():
@@ -435,6 +449,26 @@ def integrity_iter(iter, checksum):
                       (md5sum, checksum))
 
 
+def serious_integrity_iter(iter, hasher, hash_value):
+    """Check image data integrity using the Glance "multihash".
+
+    :param iter: iterable containing the image data
+    :param hasher: a hashlib object
+    :param hash_value: hexdigest of the image data
+    :raises: IOError if the hashdigest of the data is not hash_value
+    """
+    for chunk in iter:
+        yield chunk
+        if isinstance(chunk, six.string_types):
+            chunk = six.b(chunk)
+        hasher.update(chunk)
+    computed = hasher.hexdigest()
+    if computed != hash_value:
+        raise IOError(errno.EPIPE,
+                      'Corrupt image download. Hash was %s expected %s' %
+                      (computed, hash_value))
+
+
 def memoized_property(fn):
     attr_name = '_lazy_once_' + fn.__name__
 
@@ -490,5 +524,88 @@ class IterableWithLength(object):
     def next(self):
         return next(self.iterable)
 
+    # In Python 3, __next__() has replaced next().
+    __next__ = next
+
     def __len__(self):
         return self.length
+
+
+class RequestIdProxy(wrapt.ObjectProxy):
+    def __init__(self, wrapped):
+        # `wrapped` is a tuple: (original_obj, response_obj)
+        super(RequestIdProxy, self).__init__(wrapped[0])
+        self._self_wrapped = wrapped[0]
+        req_id = _extract_request_id(wrapped[1])
+        self._self_request_ids = [req_id]
+
+    @property
+    def request_ids(self):
+        return self._self_request_ids
+
+    @property
+    def wrapped(self):
+        return self._self_wrapped
+
+    # Overriden next method to act as iterator
+    def next(self):
+        return next(self._self_wrapped)
+
+    # In Python 3, __next__() has replaced next().
+    __next__ = next
+
+
+class GeneratorProxy(wrapt.ObjectProxy):
+    def __init__(self, wrapped):
+        super(GeneratorProxy, self).__init__(wrapped)
+        self._self_wrapped = wrapped
+        self._self_request_ids = []
+
+    def _set_request_ids(self, resp):
+        if self._self_request_ids == []:
+            req_id = _extract_request_id(resp)
+            self._self_request_ids = [req_id]
+
+    def _next(self):
+        obj, resp = next(self._self_wrapped)
+        self._set_request_ids(resp)
+        return obj
+
+    # Override generator's next method to add
+    # request id on each iteration
+    def next(self):
+        return self._next()
+
+    # For Python 3 compatibility
+    def __next__(self):
+        return self._next()
+
+    def __iter__(self):
+        return self
+
+    @property
+    def request_ids(self):
+        return self._self_request_ids
+
+    @property
+    def wrapped(self):
+        return self._self_wrapped
+
+
+def add_req_id_to_object():
+    @wrapt.decorator
+    def inner(wrapped, instance, args, kwargs):
+        return RequestIdProxy(wrapped(*args, **kwargs))
+    return inner
+
+
+def add_req_id_to_generator():
+    @wrapt.decorator
+    def inner(wrapped, instance, args, kwargs):
+        return GeneratorProxy(wrapped(*args, **kwargs))
+    return inner
+
+
+def _extract_request_id(resp):
+    # TODO(rsjethani): Do we need more checks here?
+    return resp.headers.get('x-openstack-request-id')
